#ifndef CLOTHOID_H
#define CLOTHOID_H

#include <QtOpenGL>
#include <cmath>

#define SAMPLES_PER_UNIT 1.0f

class Clothoid
{

public:

    Clothoid();

    void DrawGL() const;
    void DrawArcLengthGL(const float s1, const float s2) const;

    void DoSamples();

    void SetArcLen(const float f);
    void SetStartCurv(const float f);
    void SetEndCurv(const float f);

    float ArcLen() const;
    float StartCurv() const;
    float EndCurv() const;

    QVector2D Sample(int t) const;

    float Curvature(float t) const;

    QVector2D EndPoint() const;
    float EndTanAngle() const;

    //to debug with qt
    friend QDebug operator<<(QDebug d, const Clothoid & c);

private:

    float ComputeB() const;
    float ComputeT1(const float B) const;
    float ComputeT2(const float B) const;

    static QVector2D Fresnel(float s);

    static float CURV_FACTOR;
    static float SQRT_2;
    static float MIN_CURV_DIFF;
    static float MIN_CURV_CIRCLE;

    float m_startCurv;
    float m_endCurv;
    float m_arcLen;

    QVector <QVector2D> m_samples;

};

#endif // CLOTHOID_H
