#ifndef CLOTHOIDSPLINE_H
#define CLOTHOIDSPLINE_H

#include <QtCore>
#include <QtOpenGL>

#include "clothoid.h"
#include "polyline.h"
#include "util.h"

class ClothoidSpline
{

public:

    ClothoidSpline();

    void Add();
    void Add(Clothoid & c);
    void Add(ClothoidSpline & cs);
    void Remove();
    int Size() const;
    float Length() const;
    void Clear();
    const Clothoid & operator[](const int ind) const;
    Clothoid & operator[](const int ind);

    void DrawColours(const QVector3D col1, const QVector3D col2);
    void DrawArcLengthGL(const float s1, const float s2) const;
    void DrawGL(const bool closed) const;
    void DrawGL(const int start, const int end, const bool closed) const;
    void DrawTanSpaceGL(const int viewx, const int viewy, const int vieww, const int viewh, const float arclen) const;
    void DrawCurvSpaceGL(const int viewx, const int viewy, const int vieww, const int viewh, const float arclen) const;

    void SetStartPosition(const QVector2D & p);
    QVector2D StartPosition() const;
    void SetStartTangent(const QVector2D & t);
    void SetStartTangent(const float t);
    float StartTangent() const;

    QVector2D CanonEndPosition(const int ind) const;
    QVector2D CanonEndPosition() const;
    float CanonEndTangent(const int ind) const;
    float CanonEndTangent() const;

    QVector2D Position(const float t, float & r) const;

    QVector2D EndPosition(const int ind) const;
    QVector2D EndPosition() const;
    float EndTangent(const int ind) const;
    float EndTangent() const;

    void FromPolyline(const PolyLine & p, const bool forceFit);

    void CurvatureSamples(const float incr, QVector <float> & curv) const;

    float AlignToEndpoints(const QVector2D & p1, const QVector2D & p2);

    void StartTrail(const int numc, ClothoidSpline & newcs) const;
    void EndTrail(const int numc, ClothoidSpline & newcs) const;
    void Blend(const ClothoidSpline &cs, const int blendwin, QVector <QVector2D> & blendpts) const;

    static QVector2D Rotate(const QVector2D & pt,  const float angle);

private:   

    static float InterpolateG2(float f);

    QVector <float> m_lens;    
    QVector <float> m_curv;

    QVector <Clothoid> m_clothoids;

    QVector3D m_col1;
    QVector3D m_col2;

    QVector2D m_startPos;
    float m_startTan;
    float m_length;

};

#endif // CLOTHOIDSPLINE_H
