#include "frenchcurve.h"

FrenchCurve::FrenchCurve() :
        m_maxArcLen(0.0f),
        m_blend(true),
        m_blendwin(10),
        m_pointspaceerr(15.0f)
{

    m_colours.push_back(QVector3D(255, 51, 102));
    m_colours.push_back(QVector3D(255, 102, 51));
    m_colours.push_back(QVector3D(255, 204, 51));
    m_colours.push_back(QVector3D(204, 255, 51));
    m_colours.push_back(QVector3D(102, 255, 51));
    m_colours.push_back(QVector3D(51, 255, 204));

    m_colours.push_back(QVector3D(51, 204, 255));
    m_colours.push_back(QVector3D(51, 102, 255));
    m_colours.push_back(QVector3D(204, 51, 255));
    m_colours.push_back(QVector3D(255, 51, 204));

    m_colours.push_back(QVector3D(184, 138, 0));
    m_colours.push_back(QVector3D(0, 46, 184));

    for (int i=0; i<m_colours.size(); ++i)
        m_colours[i] *= (1.0f / 255.0f);

}

void FrenchCurve::Add(ClothoidSpline & c)
{

    m_maxArcLen = qMax(m_maxArcLen, c.Length());

    m_csplines.push_back(c);

    m_csplines.last().DrawColours(QVector3D(0.75, 0.75, 0.75), QVector3D(0.75, 0.75, 0.75));
    //m_csplines.last().DrawColours(m_colours[m_csplines.size()-1], m_colours[m_csplines.size()-1]);

    QVector <float> cscurv;
    m_csplines.last().CurvatureSamples(1.0f, cscurv);
    m_cscurv.push_back(cscurv);

    //qDebug() << "mscurv" << m_cscurv;

}

bool FrenchCurve::Empty()
{
    return m_csplines.empty();
}

void FrenchCurve::DrawGL() const
{


    glLineWidth(4.0f);
    for (int i=0; i<m_csplines.size(); ++i)
        m_csplines[i].DrawGL(false);

    QVector <int> shifts(m_parts.size(), 1);
    for (int i=0; i<m_parts.size(); ++i) {

        glMatrixMode(GL_MODELVIEW);
        glPushMatrix();

        glTranslatef(shifts[m_parts[i].ind]*4, shifts[m_parts[i].ind]*4, 0.0f);
        ++shifts[m_parts[i].ind];

        int cind = i % m_colours.size();
        glLineWidth(2.0f);
        glColor3f(m_colours[cind].x(), m_colours[cind].y(), m_colours[cind].z());

        m_csplines[m_parts[i].ind].DrawArcLengthGL(m_parts[i].t1, m_parts[i].t2);

        glPopMatrix();

    }

    /*
    glColor3f(0.25f, 0.75f, 0.75f);
    glPushMatrix();
    glTranslatef(250, 400, 0);
    for (int i=0; i<m_plines.size(); ++i) {
        m_plines[i].DrawGL();
    }
    glPopMatrix();
    */


    //draw french curve overlays
    /*
    for (int i=0; i<m_parts.size(); ++i) {

        const int ind = m_parts[i].ind;

        QVector2D v1 = m_parts[i].cspline.StartPosition();
        QVector2D v2 = m_parts[i].cspline.EndPosition();
        QVector2D v2m1 = v2 - v1;
        QVector2D v2m1n = v2m1.normalized();
        QVector2D vmid = (v1 + v2) * 0.5f;
        const float vtan = atan2(v2m1.y(), v2m1.x());

        float r1, r2;
        QVector2D p1 = m_csplines[ind].Position(m_parts[i].t1, r1);
        QVector2D p2 = m_csplines[ind].Position(m_parts[i].t2, r2);
        QVector2D p2m1 = p2 - p1;
        QVector2D p2m1n = (p2 - p1).normalized();
        QVector2D pmid = (p1 + p2) * 0.5f;
        const float ptan = atan2(p2m1.y(), p2m1.x());

        //const float error = p2m1.length() - cp.length();
        //SetStartPosition(p1 + p2m1n * error / 2.0f);
        //SetStartTangent(ptan - cptan);

        QVector2D startPos = vmid - pmid;

        qDebug() << vmid << pmid << startPos << vtan << ptan;

        glLineWidth(3.0f);
        glPushMatrix();

            glTranslatef(vmid.x(), vmid.y(), 0.0f);

            //if it's flipped there needs to be a mirror in tangent dir
            if (m_parts[i].flip) {
                glScalef(-1,1,1);
                glRotatef(180.0f + (vtan - ptan) * 180.0f / PI, 0, 0, 1);
            }
            else
                glRotatef((vtan - ptan) * 180.0f / PI, 0, 0, 1);

            glTranslatef(-pmid.x(), -pmid.y(), 0.0f);

            m_csplines[ind].DrawGL(true);


            glBegin(GL_LINES);
                glColor3f(1,0,0);
                glVertex2f(0, 0);
                glColor3f(0,1,0);
                glVertex2f(100, 0);
                glColor3f(1,0,0);
                glVertex2f(0, 0);
                glColor3f(0,0,1);
                glVertex2f(0, 100);
            glEnd();


            glBegin(GL_LINES);
                glVertex2f(p1.x(), p1.y());
                glVertex2f(p2.x(), p2.y());
            glEnd();


        glPopMatrix();


    }
    glLineWidth(5.0f);
    */

    glLineWidth(4.0f);
    for (int i=0; i<m_parts.size(); ++i) {

        int range1 = (i>0) ? qMin(qMin(m_blendwin, m_parts[i].cspline.Size() / 2), m_parts[i-1].cspline.Size() / 2) : 0;
        int range2 = (i<m_parts.size()-1) ? qMin(qMin(m_blendwin, m_parts[i].cspline.Size() / 2), m_parts[i+1].cspline.Size() / 2) : 0;

        if (m_curveclosed) {
            if (i==0)
                range1 = qMin(qMin(m_blendwin, m_parts.first().cspline.Size() / 2), m_parts.last().cspline.Size() / 2);
            if (i==m_parts.size()-1)
                range2 = qMin(qMin(m_blendwin, m_parts.first().cspline.Size() / 2), m_parts.last().cspline.Size() / 2);
        }

        if (m_blend)
            m_parts[i].cspline.DrawGL(range1, m_parts[i].cspline.Size() - range2, false);
        else
            m_parts[i].cspline.DrawGL(false);
    }

    for (int i=0; i<m_blendpts.size(); ++i) {
        glColor3f(0.75f, 0.75f, 0.75f);
        glBegin(GL_LINE_STRIP);
        for (int j=0; j<m_blendpts[i].size(); ++j) {
            const QVector2D & p = m_blendpts[i][j];
            glVertex2f(p.x(), p.y());

        }
        glEnd();
    }

}

void FrenchCurve::DrawCurvGL() const
{

    int winy = 10;
    for (int i=0; i<m_csplines.size(); ++i) {

        //draw each cspline curv plot
        m_csplines[i].DrawCurvSpaceGL(10, winy, 400, 100, m_maxArcLen);

        //draw part overlay
        for (int j=0; j<m_parts.size(); ++j) {

            if (m_parts[j].ind == i)
                DrawPartGL(j, 10, winy, 400, 100, m_maxArcLen);
        }

        //move up viewport
        winy += 100 + 10;
    }

}

void FrenchCurve::DrawOutCurvGL(const int viewx, const int viewy, const int vieww, const int viewh, const float arclen) const
{

    if (m_parts.empty())
        return;

    //temporarily modify opengl viewport/projection matrix
    const double curvRange = 0.05f;
    Util::PushDrawOrthoGL(viewx, viewy, vieww, viewh, 0.0, qMax(arclen, 1.0f), -curvRange, curvRange * 2.0);

    //axis lines
    glColor3f(0, 0, 0);
    glLineWidth(2.0f);

    glBegin(GL_LINES);

    glVertex2f(0.1f, -curvRange);
    glVertex2f(0.1f, curvRange);

    glVertex2f(0, 0);
    glVertex2f(arclen, 0);

    float curArc = 0.0f;
    for (int i=0; i<m_parts.size(); ++i) {

        if (i%2 == 0)
            glColor3f(1,.5,.5);
        else
            glColor3f(.5,.5,1);

        const ClothoidSpline & c = m_parts[i].cspline;

        for (int j=0; j<c.Size(); ++j) {
            glVertex2f(curArc, c[j].StartCurv());
            curArc += c[j].ArcLen();
            glVertex2f(curArc, c[j].EndCurv());
        }

    }

    glEnd();

    glLineWidth(1.0f);

    Util::PopDrawOrthoGL();
}

void FrenchCurve::DrawPartGL(const int p, const int viewx, const int viewy, const int vieww, const int viewh, const float arclen) const
{

    //temporarily modify opengl viewport/projection matrix
    Util::PushDrawOrthoGL(viewx, viewy, vieww, viewh, 0, arclen, 0, viewh);

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    if (!m_parts[p].flip)
        glColor4f(0, 1, 0, 0.2f);
    else
        glColor4f(0, 0, 1, 0.2f);

    const float x = ceil(3.0f * arclen / float(vieww));
    const float y = 3.0f;

    glBegin(GL_QUADS);

    glVertex2f(m_parts[p].t1, (p+1)*y);
    glVertex2f(m_parts[p].t2, (p+1)*y);
    glVertex2f(m_parts[p].t2, viewh);
    glVertex2f(m_parts[p].t1, viewh);

    glEnd();

    glDisable(GL_BLEND);

    glColor3f(0, 0, 0);
    glLineWidth(1.0f);

    glBegin(GL_LINES);
    glVertex2f(m_parts[p].t1, (p+1)*y);
    glVertex2f(m_parts[p].t2, (p+1)*y);
    glEnd();

    //arrow head
    if (!m_parts[p].flip) {        
        glBegin(GL_TRIANGLES);
        glVertex2f(m_parts[p].t2-x, (p+1)*y+3);
        glVertex2f(m_parts[p].t2-x, (p+1)*y-3);
        glVertex2f(m_parts[p].t2, (p+1)*y);
        glEnd();
    }
    else {        
        glBegin(GL_TRIANGLES);
        glVertex2f(m_parts[p].t1+x, (p+1)*y-3);
        glVertex2f(m_parts[p].t1+x, (p+1)*y+3);
        glVertex2f(m_parts[p].t1, (p+1)*y);
        glEnd();
    }

    Util::PopDrawOrthoGL();

}

void FrenchCurve::SetECost(const float f)
{
    m_ecost = f;
}

void FrenchCurve::SetPointSpaceError(const float f)
{
    m_pointspaceerr = f;
}

void FrenchCurve::SetBlendPieces(const bool doblend, const int blendwin)
{

    m_blend = doblend;
    m_blendwin = blendwin;

}

void FrenchCurve::Fit_Greedy(const PolyLine & pline, ClothoidSpline & incs, ClothoidSpline & outcs)
{

    incs.FromPolyline(pline, false);

    QVector <float> inccurv;
    incs.CurvatureSamples(1.0f, inccurv);

    QVector <float> arclens;
    float curlen = 0.0f;
    for (int i=0; i<incs.Size(); ++i) {
        arclens.push_back(curlen);
        curlen+=incs[i].ArcLen();
    }
    arclens.push_back(curlen);

    const int sz = incs.Size() + 1;
    m_parts.clear();

    for (int i=0; i<sz-1; ++i) {

        int j;
        float lastcost = FLT_MAX;

        for (j=i+1; j<sz; ++j) {

            //we need to convert curvature-space verts into discretely sampled range
            const int i1 = int(arclens[i]);
            const int i2 = int(arclens[j]);

            //get best fit
            float fiterr;
            int fitcscurvind;

            FrenchCurvePart newp;

            FindBestLocalFit(inccurv, i1, i2, fiterr, newp.ind, fitcscurvind, newp.flip);
            //fiterr += m_ecost;

            newp.i1 = i;
            newp.i2 = j;
            newp.t1 = float(fitcscurvind);
            newp.t2 = newp.t1 + (arclens[j] - arclens[i]);

            if (!newp.flip) {
                for (float f=newp.t1; f<newp.t2-1.0f; f += 1.0f) {

                    int ci = int(f);

                    Clothoid c;
                    c.SetArcLen(1.0f);
                    c.SetStartCurv(m_cscurv[newp.ind][ci]);
                    c.SetEndCurv(m_cscurv[newp.ind][ci+1]);

                    c.DoSamples();

                    newp.cspline.Add(c);

                }
            }
            else {
                for (float f=newp.t2; f>newp.t1+1.0f; f -= 1.0f) {

                    int ci = int(f);

                    Clothoid c;
                    c.SetArcLen(1.0f);
                    c.SetStartCurv(-m_cscurv[newp.ind][ci]);
                    c.SetEndCurv(-m_cscurv[newp.ind][ci-1]);

                    c.DoSamples();

                    newp.cspline.Add(c);

                }
            }

            //match each piece's endpoints as well as possible
            const float spacefiterr = newp.cspline.AlignToEndpoints(pline[i], pline[j]);

            if ((fiterr > lastcost + m_ecost || spacefiterr > 10.0f) && i2 > i1 + 1) {
                --j;
                break;
            }
            else {

                lastcost = fiterr;

                //replace part i...j-1 with part i...j
                if (!m_parts.empty() && m_parts.last().i1 == i) {
                    //qDebug()<<"removing part for range" <<m_parts.last().i1<<"to" <<m_parts.last().i2;
                    m_parts.pop_back();
                }

                //qDebug()<<"adding part for range " << i << "to" << j<<"sz is" <<sz;
                m_parts.push_back(newp);
            }

        }

        i=j-1;

    }

    for (int i=0; i<m_parts.size(); ++i) {

        //set colour
        if (i % 2 == 0)
            m_parts[i].cspline.DrawColours(QVector3D(1,.5,.5), QVector3D(1,.5,.5));
        else
            m_parts[i].cspline.DrawColours(QVector3D(.5,.5,1), QVector3D(.5,.5,1));

        outcs.Add(m_parts[i].cspline);
    }

    if (m_blend)
        DoBlendPieces();

}

void FrenchCurve::Fit(const PolyLine & pline, ClothoidSpline & incs, ClothoidSpline & outcs)
{

    const float endpointlen = (pline[0] - pline[pline.Size()-1]).length();
    m_curveclosed = endpointlen < 40.0f ? true : false;

    incs.FromPolyline(pline, false);

    QVector <float> inccurv;
    incs.CurvatureSamples(1.0f, inccurv);

    QVector <float> arclens;
    float curlen = 0.0f;
    for (int i=0; i<incs.Size(); ++i) {
        arclens.push_back(curlen);
        curlen+=incs[i].ArcLen();
    }
    arclens.push_back(curlen);

    //populate each diagonal of matrix M
    //entries of M are for each point of the input curve (and its associated curvature space points)
    const int sz = incs.Size() + 1; //number of *points* in curvature space
    QVector <QVector <MatrixEntry> > M(sz, QVector <MatrixEntry> (sz));

    for (int i=0; i<sz; ++i) {
        for (int j=i; j<sz; ++j) {
            M[i][j].row = i;
            M[i][j].col = j;
            M[i][j].cost = m_ecost;
            M[i][j].splitindex = -1;
        }
    }

    for (int i=0; i<sz; ++i) { //each diagonal

        for (int j=0; j+i+1<sz; ++j) { //each element of each diagonal

            const int vi1 = j;
            const int vi2 = j + i + 1;

            //we need to convert curvature-space verts into discretely sampled range
            const int i1 = int(arclens[vi1]);
            const int i2 = int(arclens[vi2]);

            //get best fit
            float fiterr = FLT_MAX;
            int fitcsind = 0;
            int fitcssampleind = 0;
            bool fitflip = false;

            FindBestLocalFit(inccurv, i1, i2, fiterr, fitcsind, fitcssampleind, fitflip);
            fiterr += m_ecost;

            //get best split
            float spliterr = FLT_MAX;
            int splitind = 0;

            for (int split=vi1+1; split<vi2; ++split) {

                if (M[vi1][split].cost + M[split][vi2].cost < spliterr) {
                    spliterr = M[vi1][split].cost + M[split][vi2].cost;
                    splitind = split;
                }

            }

            //ALWAYS write the best french curve info for each sub curve of input curve (as there may be a post-process split)
            M[vi1][vi2].fcsind = fitcsind;
            M[vi1][vi2].fcssampleind = fitcssampleind;
            M[vi1][vi2].fcsflip = fitflip;

            //split/choose cost based on what's optimal
            if (fiterr < spliterr) {
                M[vi1][vi2].cost = fiterr;
                M[vi1][vi2].splitindex = -1;                
            }
            else {
                M[vi1][vi2].cost = spliterr;
                M[vi1][vi2].splitindex = splitind;
            }

        }

    }

    QVector <MatrixEntry> parts;
    RecurseMatrix(M, 0, sz-1, parts);

    m_parts.clear();

    for (int i=0; i<parts.size(); ++i) {

        QVector <FrenchCurvePart> p;
        ConstructSplineForRange(M, parts[i].row, parts[i].col, pline, arclens, p);

        for (int j=0; j<p.size(); ++j) {
            outcs.Add(p[j].cspline);
            m_parts.push_back(p[j]);
        }

    }

    //set spline part colours
    for (int i=0; i<m_parts.size(); ++i) {

        /*
        if (i % 2 == 0)
            m_parts[i].cspline.DrawColours(QVector3D(1,.5,.5), QVector3D(1,.5,.5));
        else
            m_parts[i].cspline.DrawColours(QVector3D(.5,.5,1), QVector3D(.5,.5,1));
            */

        int cind = i % m_colours.size();
        m_parts[i].cspline.DrawColours(m_colours[cind], m_colours[cind]);

    }

    if (m_blend)
        DoBlendPieces();

    qDebug() << "Curvature error in French curve fit: " << M[0][sz-1].cost << " using " << parts.size() << "pieces.";

    //now we walk through the matrix and construct the clothoid spline out of pieces of this curve
    /*
    qDebug () << M;
    qDebug () << W;
    qDebug () << F;
    qDebug() << "parts: " << parts;
    */    

}

void FrenchCurve::DoBlendPieces()
{

    m_blendpts.clear();   

    if (m_parts.empty())
        return;

    for (int i=0; i<m_parts.size()-1; ++i) {

        const ClothoidSpline & cs1 = m_parts[i].cspline;
        const ClothoidSpline & cs2 = m_parts[i+1].cspline;        

        QVector <QVector2D> blendpts;
        cs1.Blend(cs2, m_blendwin,  blendpts);
        m_blendpts.push_back(blendpts);

    }  

    if (m_curveclosed && !m_parts.empty()) {
        qDebug() << "closing curve!";
        const ClothoidSpline & cs1 = m_parts.last().cspline;
        const ClothoidSpline & cs2 = m_parts.first().cspline;

        QVector <QVector2D> blendpts;
        cs1.Blend(cs2, m_blendwin,  blendpts);
        m_blendpts.push_back(blendpts);
    }

}


void FrenchCurve::ConstructSplineForRange(const QVector <QVector <MatrixEntry> > & M, const int i1, const int i2, const PolyLine & pline, const QVector <float> & arclens, QVector <FrenchCurvePart> & p)
{

    FrenchCurvePart newp;

    newp.ind = M[i1][i2].fcsind;
    newp.t1 = M[i1][i2].fcssampleind;
    newp.t2 = newp.t1 + int(arclens[i2] - arclens[i1]);
    newp.flip = M[i1][i2].fcsflip;

    if (!newp.flip) {

        for (int ci=newp.t1; ci<newp.t2; ++ci) {

            Clothoid c;
            c.SetArcLen(1.0f);
            c.SetStartCurv(m_cscurv[newp.ind][ci]);
            c.SetEndCurv(m_cscurv[newp.ind][ci+1]);

            c.DoSamples();

            newp.cspline.Add(c);           
        }
    }
    else {
        for (int ci=newp.t2; ci>newp.t1; --ci) {

            Clothoid c;
            c.SetArcLen(1.0f);
            c.SetStartCurv(-m_cscurv[newp.ind][ci]);
            c.SetEndCurv(-m_cscurv[newp.ind][ci-1]);

            c.DoSamples();

            newp.cspline.Add(c);

        }
    }

    //match each piece's endpoints as well as possible
    const float fiterr = newp.cspline.AlignToEndpoints(pline[i1], pline[i2]);

    if (fiterr > m_pointspaceerr && i2 > i1 + 1) {

        //qDebug() << "fiterror " << fiterr << "between pts " << i1 << "to" << i2 << "midpt " << midpt << " so recursing.";

        int bestpt = -1;
        float bestcost = FLT_MAX;

        for (int i=i1+1; i<i2; ++i) {

            const float eachcost = M[i1][i].cost + M[i][i2].cost;

            if (eachcost < bestcost) {
                bestcost = eachcost;
                bestpt = i;
            }

        }

        ConstructSplineForRange(M, i1, bestpt, pline, arclens, p);
        ConstructSplineForRange(M, bestpt, i2, pline, arclens, p);

    }
    else {
        p.push_back(newp);
    }


}

void FrenchCurve::RecurseMatrix(const QVector <QVector <MatrixEntry> > & M, const int r1, const int r2, QVector <MatrixEntry> & parts) const
{

    if (M[r1][r2].splitindex == -1) { //splitindex is -1 - indicates piece covers whole range
        parts.push_back(M[r1][r2]);
    }
    else { //entry is non negative, it's a split index
        RecurseMatrix(M, r1, M[r1][r2].splitindex, parts);
        RecurseMatrix(M, M[r1][r2].splitindex, r2, parts);
    }

}

void FrenchCurve::FindBestLocalFit(const QVector <float> & inccurv, const int i1, const int i2, float & err, int & csind, int & cssampleind, bool & flip) const
{

    //sweep along a window of size i1-i2 each curve in the french curve set
    int window = i2 - i1 + 1; //inclusive range

    //need a function that compares the area between two polylines (where both DO span an interval of size "window")
    err = FLT_MAX;
    csind = 0;
    cssampleind = 0;
    flip = false;

    //for each curve
    for (int i=0; i<m_cscurv.size(); ++i) {

        //slide window along
        for (int j=0; j+window<m_cscurv[i].size(); ++j) {

            float eacherr = 0.0f;
            float int1 = 0.0f;
            float int2 = 0.0f;

            bool computeint = false;

            if (!computeint) {

                //compute difference for window
                for (int k=0; k<window; ++k)
                    eacherr += fabs(inccurv[i1+k] - m_cscurv[i][j+k]);
            }
            else {

                //integrate curvature as move along way (good for G1 discont)

                for (int k=0; k<window; ++k) {
                    int1 += inccurv[i1+k];
                    int2 += m_cscurv[i][j+k];
                    eacherr += fabs(int1 - int2);
                }
            }

            //if difference minimal, save this position
            if (eacherr < err) {
                err = eacherr;
                csind = i;
                cssampleind = j;
                flip = false;
            }


            //do again, but flip direction
            eacherr = 0.0f;
            int1 = 0.0f;
            int2 = 0.0f;

            if (!computeint) {

                for (int k=0; k<window; ++k)
                    eacherr += fabs(inccurv[i1+k] + m_cscurv[i][j+(window-1)-k]);

            }
            else {

                for (int k=0; k<window; ++k) {
                    int1 += inccurv[i1+k];
                    int2 -= m_cscurv[i][j+(window-1)-k];
                    eacherr += fabs(int1 - int2);
                }

            }

            //if difference minimal, save this position
            if (eacherr < err) {
                err = eacherr;
                csind = i;
                cssampleind = j;
                flip = true;
            }

        }

    }

}

