#ifndef FRENCHCURVE_H
#define FRENCHCURVE_H

#include <QtGui>
#include <cfloat>

#include "clothoidspline.h"
#include "util.h"

struct FrenchCurvePart {

    int i1;
    int i2;

    int ind;
    int t1;
    int t2;
    bool flip;

    ClothoidSpline cspline;
};

struct MatrixEntry {

    int row;
    int col;
    float cost;

    int splitindex;
    int fcsind;
    int fcssampleind;
    bool fcsflip;

};

class FrenchCurve
{

public:

    FrenchCurve();

    void Add(ClothoidSpline & c);
    bool Empty();
    void DrawGL() const;
    void DrawCurvGL() const;    
    void DrawOutCurvGL(const int viewx, const int viewy, const int vieww, const int viewh, const float arclen) const;

    void Fit(const PolyLine & pline, ClothoidSpline & incs, ClothoidSpline & outcs);
    void Fit_Greedy(const PolyLine & pline, ClothoidSpline & incs, ClothoidSpline & outcs);

    void SetECost(const float f);
    void SetPointSpaceError(const float f);

    void SetBlendPieces(const bool doblend, const int blendwin);
    void DoBlendPieces();

private:

    void BlendClothoids(const ClothoidSpline & cs1, const ClothoidSpline & cs2, QVector <QVector2D> & blendpts);

    void DrawPartGL(const int p, const int viewx, const int viewy, const int vieww, const int viewh, const float arclen) const;

    void ConstructSplineForRange(const QVector <QVector <MatrixEntry> > & M, const int i1, const int i2, const PolyLine & pline, const QVector <float> & arclens, QVector <FrenchCurvePart> & p);
    void FindBestLocalFit(const QVector <float> & inccurv, const int i1, const int i2, float & err, int & csind, int & cssampleind, bool & flip) const;
    void FindBestLocalFit_TanAngle(const QVector <float> & inccurv, const int i1, const int i2, float & err, int & csind, int & cssampleind, bool & flip) const;
    void RecurseMatrix(const QVector <QVector <MatrixEntry> > & M, const int r1, const int r2, QVector <MatrixEntry> & parts) const;    

    QVector <ClothoidSpline> m_csplines;    
    QVector <QVector <float> > m_cscurv;
    QVector <FrenchCurvePart> m_parts;

    //sets of points which smoothly interpolate between french curve parts
    QVector <QVector <QVector2D> > m_blendpts;

    QVector <QVector3D> m_colours;

    float m_maxArcLen;
    float m_ecost;

    bool m_blend;
    int m_blendwin;

    bool m_curveclosed;
    float m_pointspaceerr;

};

#endif // FRENCHCURVE_H
