#pragma once

#include <QtGui>
#include <QtOpenGL>

#include "frenchcurve.h"
#include "clothoidspline.h"
#include "polyline.h"

class GLWidget : public QGLWidget
{

    Q_OBJECT

public:
    GLWidget(QWidget *parent = 0);
    ~GLWidget();

    QSize minimumSizeHint() const;
    QSize sizeHint() const;

public slots:

    void ChangedECost(int ecost);
    void ChangedBlendWin(int blendwin);
    void ChangedPointError(int pointerr);

    void SetDrawPolyLine(bool b);
    void SetDrawFrenchCurves(bool b);
    void SetDoGreedy(bool b);

protected:

    void initializeGL();
    void paintGL();    
    void resizeGL(int width, int height);

    void mousePressEvent(QMouseEvent * e);
    void mouseMoveEvent(QMouseEvent * e);
    void mouseReleaseEvent(QMouseEvent * e);
    void keyReleaseEvent(QKeyEvent * e);

private:

    void DoFrenchCurveFit();

    int winWidth, winHeight;
    int mouse_x, mouse_y;

    FrenchCurve fcurve;

    PolyLine pline;
    ClothoidSpline cspline;
    ClothoidSpline ocspline;        

    bool m_bdrawpoly;
    bool m_bdrawfrench;
    bool m_dogreedy;

};

